/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_ngraph.c

#include "quakedef.h"
#include "shader.h"

extern qbyte		*draw_chars;				// 8*8 graphic characters

static texid_t	netgraphtexture;	// netgraph texture
static shader_t *netgraphshader;

static int timehistory[NET_TIMINGS];
static int findex;

#define NET_GRAPHHEIGHT 32

static	qbyte ngraph_texels[NET_GRAPHHEIGHT][NET_TIMINGS];

static void R_LineGraph (int x, int h)
{
	int		i;
	int		s;
	int		color, color2 = 0xff;

	s = NET_GRAPHHEIGHT;

	if (h == 10000 || h<0)
	{
		color = 0;	// yellow
		color2 = 1;
		h=abs(h);
	}
	else if (h == 9999)
	{
		color = 2;	// red
		color2 = 3;
	}
	else if (h == 9998)
	{
		color = 4;	// blue
		color2 = 5;
	}
	else
	{
		color = 6;	// white
		color2 = 7;
	}

	if (h>s)
		h = s;
	
	for (i=0 ; i<h ; i++)
		if (i & 1)
			ngraph_texels[NET_GRAPHHEIGHT - i - 1][x] = (qbyte)color2;
		else
			ngraph_texels[NET_GRAPHHEIGHT - i - 1][x] = (qbyte)color;

	for ( ; i<s ; i++)
		ngraph_texels[NET_GRAPHHEIGHT - i - 1][x] = (qbyte)8;
}

/*
static void Draw_CharToNetGraph (int x, int y, int num)
{
	int		row, col;
	qbyte	*source;
	int		drawline;
	int		nx;

	if (!draw_chars)
		return;

	row = num>>4;
	col = num&15;
	source = draw_chars + (row<<10) + (col<<3);

	for (drawline = 8; drawline; drawline--, y++)
	{
		for (nx=0 ; nx<8 ; nx++)
			if (source[nx] != 255)
				ngraph_texels[y][nx+x] = 0x60 + source[nx];
		source += 128;
	}
}
*/

/*
==============
R_NetGraph
==============
*/
//instead of assuming the quake palette, we should use a predictable lookup table. it makes the docs much easier.
static unsigned ngraph_palette[] =
{
	0xff00ffff,	//yellow
	0xff00efef,	//yellow2
	0xff0000ff,	//red
	0xff0000ef,	//red2
	0xffff0000,	//blue
	0xffef0000,	//blue2
	0xffffffff,	//white
	0xffefefef,	//white2
	0x00000000	//invisible.
};
void R_NetGraph (void)
{
	int		a, x, i, y;
	int lost;
	char st[80];
	unsigned	ngraph_pixels[NET_GRAPHHEIGHT][NET_TIMINGS];
	float pi, po, bi, bo;

	x = 0;
	if (r_netgraph.value < 0)
	{
		lost = -1;
		if (!cl.paused)
			timehistory[++findex&NET_TIMINGSMASK] = (cl.currentpackentities?(cl.currentpackentities->servertime - cl.servertime)*NET_GRAPHHEIGHT*5:0);
		for (a=0 ; a<NET_TIMINGS ; a++)
		{
			i = (findex-a) & NET_TIMINGSMASK;
			R_LineGraph (NET_TIMINGS-1-a, timehistory[i]<0?10000:timehistory[i]);
		}
	}
	else
	{
		int last = 10000;
		lost = CL_CalcNet(r_netgraph.value);
		for (a=0 ; a<NET_TIMINGS ; a++)
		{
			i = (cl.movesequence-a) & NET_TIMINGSMASK;
			if (packet_latency[i] != 10000)
				last = packet_latency[i];
			else if (last >= 0)
				last = -last;
			R_LineGraph (NET_TIMINGS-1-a, last);
		}
	}

	// now load the netgraph texture into gl and draw it
	for (y = 0; y < NET_GRAPHHEIGHT; y++)
		for (x = 0; x < NET_TIMINGS; x++)
			ngraph_pixels[y][x] = ngraph_palette[ngraph_texels[y][x]];

	x =	((vid.width - 320)>>1);
	x=-x;
	y = vid.height - sb_lines - 24 - NET_GRAPHHEIGHT - 2*8;

	M_DrawTextBox (x, y, NET_TIMINGS/8, NET_GRAPHHEIGHT/8 + 3);
	y += 8;

	sprintf(st, "%3i%% packet loss", lost);
	Draw_FunString(8, y, st);
	y += 8;

	if (NET_GetRates(cls.sockets, &pi, &po, &bi, &bo))
	{
		Draw_FunString(8, y+0, va("in: %g %g\n", pi, bi));	//not relevent as a limit.
		Draw_FunString(8, y+8, va("out: %g %g\n", po, bo));	//not relevent as a limit.
	}
	y += 16;

	Image_Upload(netgraphtexture, TF_RGBA32, ngraph_pixels, NULL, NET_TIMINGS, NET_GRAPHHEIGHT, IF_UIPIC|IF_NOMIPMAP|IF_NOPICMIP);
	x=8;
	R2D_Image(x, y, NET_TIMINGS, NET_GRAPHHEIGHT, 0, 0, 1, 1, netgraphshader);
}

void R_FrameTimeGraph (int frametime)
{
	int		a, x, i, y;
	unsigned	ngraph_pixels[NET_GRAPHHEIGHT][NET_TIMINGS];

	timehistory[findex++&NET_TIMINGSMASK] = frametime;

	x = 0;
	for (a=0 ; a<NET_TIMINGS ; a++)
	{
		i = (findex-a) & NET_TIMINGSMASK;
		R_LineGraph (NET_TIMINGS-1-a, timehistory[i]);
	}

	// now load the netgraph texture into gl and draw it
	for (y = 0; y < NET_GRAPHHEIGHT; y++)
		for (x = 0; x < NET_TIMINGS; x++)
			ngraph_pixels[y][x] = d_8to24rgbtable[ngraph_texels[y][x]];

	x =	((vid.width - 320)>>1);
	x=-x;
	y = vid.height - sb_lines - 24 - NET_GRAPHHEIGHT - 1;

	M_DrawTextBox (x, y, NET_TIMINGS/8, NET_GRAPHHEIGHT/8 + 1);
	y += 8;

	y += 8;

	Image_Upload(netgraphtexture, TF_RGBA32, ngraph_pixels, NULL, NET_TIMINGS, NET_GRAPHHEIGHT, IF_UIPIC|IF_NOMIPMAP|IF_NOPICMIP);
	x=8;
	R2D_Image(x, y, NET_TIMINGS, NET_GRAPHHEIGHT, 0, 0, 1, 1, netgraphshader);
}

void R_NetgraphInit(void)
{
	TEXASSIGN(netgraphtexture, Image_CreateTexture("***netgraph***", NULL, IF_UIPIC|IF_NOMIPMAP));
	netgraphshader = R_RegisterShader("netgraph", SUF_NONE,
		"{\n"
			"program default2d\n"
			"{\n"
				"map $diffuse\n"
				"blendfunc blend\n"
			"}\n"
		"}\n"
		);
	netgraphshader->defaulttextures->base = netgraphtexture;
}
